#!/usr/bin/perl

use strict;
use warnings;
use SNMP;
use Socket;
use Data::Dumper;
use Time::Local;
use Xorux_lib;
use Math::BigInt;
use JSON;

my $lan_ip      = $ENV{LAN_IP}   || error("LAN IP is required!")      && exit;
my $lan_type    = $ENV{LAN_TYPE} ? $ENV{LAN_TYPE} : "UNKNOWN";
my $lan_group   = $ENV{LAN_GROUP};
my $basedir     = $ENV{INPUTDIR} || error("INPUTDIR is not defined!") && exit;
my $webdir      = $ENV{WEBDIR}   ||= "$basedir/www" ;
my $wrkdir      = "$basedir/data";
my $tmpdir      = "$basedir/tmp";
my $update_conf = 0;
my %mib_compatibility;
my %data;
my %conf;

# SNMP parameters
my $dest       = $lan_ip;
my $comm       = 'public';
my $sver       = '2c';
my $timeout    = '50000000';
my $SecName    = "snmpuser1";
my $RemotePort = "";
my $switch_name;

# LAN data collection timeout
my $lan_data_collection_timeout = 600;
if ( defined $ENV{LAN_DATA_COLLECTION_TIMEOUT} && isdigit($ENV{LAN_DATA_COLLECTION_TIMEOUT}) ) {
  $lan_data_collection_timeout = $ENV{LAN_DATA_COLLECTION_TIMEOUT};
}

eval {
  # Set alarm
  my $act_time = localtime();
  local $SIG{ALRM} = sub { die "$act_time: died in SIG ALRM"; };
  alarm($lan_data_collection_timeout);

  initialize_snmp();
  check_mib_compatibility();

  my $sess = snmp_session();

  collect_data($sess);
  collect_configuration($sess);
  create_switch_directories();
  save_performance_data();
  save_configuration();
  save_health_status();

  # end of alarm
  alarm(0);
};
if ($@) {
  if ( $@ =~ /died in SIG ALRM/ ) {
    error( "Data collection via snmp timed out after $lan_data_collection_timeout seconds : " . __FILE__ . ":" . __LINE__ );
    exit(0);
  }
}

exit;


sub initialize_snmp {
  # SNMP option
  if ( defined $ENV{SNMP_VERSION} && $ENV{SNMP_VERSION} ne '' ) {
    if ( isdigit( $ENV{SNMP_VERSION} ) && $ENV{SNMP_VERSION} == 3 ) {
      $sver = $ENV{SNMP_VERSION};
    }
    elsif ( $ENV{SNMP_VERSION} eq "2c" ) {
      $sver = $ENV{SNMP_VERSION};
    }
    else {
      error( "Unknown | $ENV{SNMP_VERSION} | SNMP version in etc/lan-list.cfg! Automatically used SNMP version \"$sver\"! $!" . __FILE__ . ":" . __LINE__ );
    }
  } ## end if ( defined $ENV{SNMP_VERSION...})
  if ( defined $ENV{SNMP_PORT} && $ENV{SNMP_PORT} ne '' ) {
    $RemotePort = $ENV{SNMP_PORT};
    if ( !isdigit($RemotePort) ) {
      error( "SNMP port \"$RemotePort\" is not digit! Automatically used SNMP default port \"161\"! $!" . __FILE__ . ":" . __LINE__ );
      $RemotePort = "";
    }
    my $snmp_port_length = length $RemotePort;
    if ( isdigit($snmp_port_length) && $snmp_port_length > 5 ) {
      error( "SNMP port \"$RemotePort\" length is greater then 5! Automatically used SNMP default port \"161\"! $!" . __FILE__ . ":" . __LINE__ );
      $RemotePort = "";
    }
  } ## end if ( defined $ENV{SNMP_PORT...})

  if ( -f "$basedir/etc/lan-list.cfg" ) {
    open( SL, "<$basedir/etc/lan-list.cfg" ) || error( "Couldn't open file $basedir/etc/lan-list.cfg $!" . __FILE__ . ":" . __LINE__ );
    my @lan_cfg = <SL>;
    close(SL);
    my ($line) = grep {/^$lan_ip:/} @lan_cfg;
    chomp $line;
    my ( undef, $comm_st, undef, undef ) = split( ":", $line );
    if ( defined $comm_st && $comm_st ne "" ) {
      chomp $comm_st;
      $comm = $comm_st;
    }
    if ( defined $comm_st && $comm_st ne "" && $sver eq 3 ) {
      chomp $comm_st;
      $SecName = $comm_st;
    }
  } ## end if ( -f "$basedir/etc/lan-list.cfg")

  # Initialize the MIB (else you can't do queries).
  &SNMP::addMibDirs("$basedir/MIBs");

  &SNMP::loadModules( 'SW-MIB', 'Brocade-REG-MIB', 'Brocade-TC', 'FCMGMT-MIB', 'CISCO-FC-FE-MIB', 'CISCO-DM-MIB', 'Q-BRIDGE-MIB', 'CISCO-VTP-MIB', 'CISCO-CDP-MIB', 'IEEE8023-LAG-MIB', 'CISCO-ETHERNET-FABRIC-EXTENDER-MIB', 'CISCO-STACKWISE-MIB', 'CISCO-VIRTUAL-SWITCH-MIB', 'CISCO-PAGP-MIB', 'CISCO-ENTITY-FRU-CONTROL-MIB', 'CISCO-VLAN-IFTABLE-RELATIONSHIP-MIB', 'CISCO-MEMORY-POOL-MIB', 'CISCO-PROCESS-MIB' );
  &SNMP::initMib();

  return 1;
} ## end sub initialize_snmp

sub snmp_session {
  my $community_index = shift;
  my $params = shift;
  my %snmpparms = defined $params ? %{ $params } : ();
  my $community_string = $community_index ? "$comm\@$community_index" : $comm;
  my $sess;

  $snmpparms{Community}      = $community_string;
  $snmpparms{DestHost}       = inet_ntoa( inet_aton($dest) );
  $snmpparms{Version}        = $sver;
  $snmpparms{Timeout}        = $timeout;
  $snmpparms{UseSprintValue} = '1';
  $snmpparms{SecName}        = $SecName;

  if ( defined $RemotePort && $RemotePort ne '' ) {
    $snmpparms{RemotePort} = $RemotePort;
  }

  $sess = new SNMP::Session(%snmpparms);

  return $sess;
} ## end sub snmp_session

sub check_mib_compatibility {
  my $mib_compatibility_file = "$tmpdir/$lan_ip-mib-compatiblity.txt";

  ## check mib compatibility only once per 1 hour
  if ( !-f $mib_compatibility_file || file_time_diff( $mib_compatibility_file ) > 3600 ) {
    my $sess = snmp_session( undef, { UseLongNames => 1 } );

    $mib_compatibility{"Q-BRIDGE-MIB"} = snmp_test_mib( $sess, 'qBridgeMIB' );
    $mib_compatibility{"LLDP-MIB"}     = snmp_test_mib( $sess, 'lldpMIB' );

    if ( $lan_type eq 'CISCO' ) {
      $mib_compatibility{"CISCO-CDP-MIB"}                       = snmp_test_mib( $sess, 'ciscoCdpMIB' );
      $mib_compatibility{"CISCO-VTP-MIB"}                       = snmp_test_mib( $sess, 'ciscoVtpMIB' );
      $mib_compatibility{"CISCO-PAGP-MIB"}                      = snmp_test_mib( $sess, 'ciscoPagpMIB');
      $mib_compatibility{"CISCO-VIRTUAL-SWITCH-MIB"}            = snmp_test_mib( $sess, 'ciscoVirtualSwitchMIB');
      $mib_compatibility{"CISCO-STACKWISE-MIB"}                 = snmp_test_mib( $sess, 'ciscoStackWiseMIB');
      $mib_compatibility{"CISCO-ETHERNET-FABRIC-EXTENDER-MIB"}  = snmp_test_mib( $sess, 'ciscoEthernetFabricExtenderMIB');
      $mib_compatibility{"CISCO-PROCESS-MIB"}                   = snmp_test_mib( $sess, 'ciscoProcessMIB');
      $mib_compatibility{"CISCO-MEMORY-POOL-MIB"}               = snmp_test_mib( $sess, 'ciscoMemoryPoolMIB');
      $mib_compatibility{"CISCO-ENTITY-FRU-CONTROL-MIB"}        = snmp_test_mib( $sess, 'ciscoEntityFRUControlMIB');
      $mib_compatibility{"CISCO-VLAN-IFTABLE-RELATIONSHIP-MIB"} = snmp_test_mib( $sess, 'ciscoVlanIfTableRelationshipMIB');

      # test connection with community string index on vlan 1
      $sess = snmp_session( 1, { UseLongNames => 1 } );
      $mib_compatibility{"COMMUNITY-STRING-INDEX"}    = snmp_test_mib( $sess, 'dot1dBridge' );
    }

    save_as_json($mib_compatibility_file, \%mib_compatibility);
  }
  else {
    my $mib_compatibility_ref = read_json($mib_compatibility_file);

    if ( $mib_compatibility_ref ) {
      %mib_compatibility = %{ $mib_compatibility_ref };
    }
  }

  return 1;
}

sub collect_data {
  my $sess = shift; # The SNMP::Session object that does the work.

  snmp_collect($sess, 'configuration', 'sysDescr');
  snmp_collect($sess, 'configuration', 'sysName');

  if ( defined $data{configuration}{sysName} ) {
    $switch_name = $data{configuration}{sysName};
  }
  else {
    error( "Switch name not found! IP: $lan_ip $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }

  snmp_collect($sess, 'ports', 'ifName');
  snmp_collect($sess, 'ports', 'ifAlias');
  snmp_collect($sess, 'ports', 'ifType');
  snmp_collect($sess, 'ports', 'ifOperStatus');
  snmp_collect($sess, 'ports', 'ifAdminStatus');
  snmp_collect($sess, 'ports', 'ifHighSpeed');

  snmp_collect($sess, 'ports', 'ifHCOutOctets', \&bignum); # bytes transmitted - data out
  snmp_collect($sess, 'ports', 'ifHCInOctets', \&bignum); # bytes recived - data in
  snmp_collect($sess, 'ports', 'ifOutUcastPkts'); # Unicast Packets transmitted - Ucast_pkts_out
  snmp_collect($sess, 'ports', 'ifInUcastPkts'); # Unicast Packets recived - Ucast_pkts_in
#  snmp_collect($sess, 'ports', 'ifOutMulticastPkts'); # Multicast Packets transmitted - Mcast_pkts_out
#  snmp_collect($sess, 'ports', 'ifInMulticastPkts'); # Multicast Packets recived - Mcast_pkts_in
#  snmp_collect($sess, 'ports', 'ifOutBroadcastPkts'); # Broadcast Packets transmitted - Bcast_pkts_out
#  snmp_collect($sess, 'ports', 'ifInBroadcastPkts'); # Broadcast Packets recived - Bcast_pkts_in
  snmp_collect($sess, 'ports', 'ifOutErrors'); # transmitted Packets with Errors - errors_out
  snmp_collect($sess, 'ports', 'ifInErrors'); # recived Packets with Errors - errors_i

  if ( $lan_type eq 'CISCO' ) {
    ## switch health status
    snmp_collect($sess, 'physical_entries', 'entPhysicalDescr');
    snmp_collect($sess, 'physical_entries', 'entPhysicalSerialNum');

    if ( $mib_compatibility{'CISCO-ENTITY-FRU-CONTROL-MIB'} ) {
      snmp_collect($sess, 'switch_status', 'cefcModuleAdminStatus');
      snmp_collect($sess, 'switch_status', 'cefcModuleOperStatus');
    }
    ## fex ports
    if ( $mib_compatibility{'CISCO-ETHERNET-FABRIC-EXTENDER-MIB'} ) {
      snmp_collect($sess, 'fex_ports', 'cefexBindingExtenderIndex');
    }

    if ( $mib_compatibility{'CISCO-PROCESS-MIB'} ) {
      snmp_collect($sess, 'physical_entries', 'entPhysicalName');
      snmp_collect($sess, 'cpu', 'cpmCPUTotal5minRev');
      snmp_collect($sess, 'cpu', 'cpmCPUTotalPhysicalIndex');
    }

    if ( $mib_compatibility{'CISCO-MEMORY-POOL-MIB'} ) {
      snmp_collect($sess, 'memory', 'ciscoMemoryPoolName');
      snmp_collect($sess, 'memory', 'ciscoMemoryPoolUsed');
      snmp_collect($sess, 'memory', 'ciscoMemoryPoolFree');
    }
  }

  return 1;
} ## end sub collect_data

sub create_switch_directories {
  my $act_time = time();
  if ( !-d "$wrkdir/$switch_name" ) {
    mkdir( "$wrkdir/$switch_name", 0755 ) || error( "$act_time: Cannot mkdir $wrkdir/$switch_name: $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }
  if ( !-d "$webdir/$switch_name" ) {
    mkdir( "$webdir/$switch_name", 0755 ) || error( "$act_time: Cannot mkdir $webdir/$switch_name: $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }

  my $marker = "LAN-$lan_type";
  if ( !-f "$wrkdir/$switch_name/$marker" ) {
    open( CISC, ">$wrkdir/$switch_name/$marker" ) || error( "Couldn't open file $wrkdir/$switch_name/$marker $!" . __FILE__ . ":" . __LINE__ ) && exit;
    close(CISC);
  }

  my $group_file = "$wrkdir/$switch_name/GROUP.cfg";
  if ( defined $lan_group && $lan_group ne '' ) {
    open( GROUP, '>', $group_file ) || error( "Couldn't open file $group_file $!" . __FILE__ . ":" . __LINE__ ) && exit;
    print GROUP $lan_group;
    close(GROUP);
  }
  elsif ( -f $group_file ) {
    unlink( $group_file );
  }

  return 1;
} ## sub create_switch_directories

sub save_performance_data {
  my $act_time = time();
  my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime();
  my $date        = sprintf( "%4d%02d%02d", $year + 1900, $month + 1, $day );
  my $time        = sprintf( "%02d%02d", $hour, $min );
  my $output_file = "$wrkdir/$switch_name/$switch_name\_lanperf_$date\_$time.out.tmp\n";
  open( OUT, ">$output_file" ) || error( "Couldn't open file $output_file $!" . __FILE__ . ":" . __LINE__ ) && exit;

  #
  # PORTS.cfg
  #
  my $ports_cfg        = "$wrkdir/$switch_name/PORTS.cfg";
  my $ports_cfg_suffix = "$ports_cfg-tmp";
  open( my $pcfg, ">$ports_cfg_suffix" ) || error( "Couldn't open file $ports_cfg_suffix $!" . __FILE__ . ":" . __LINE__ ) && exit;

  print OUT "#PORT,act_time,port_id,data_out,data_in,unicast_pkts_out,unicast_pkts_in,multicast_pkts_out,multicast_pkts_in,broadcast_pkts_out,broadcast_pkts_in,errors_out,errors_in,reserve1,reserve2,switch_name,db_name\n";

  foreach my $index ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
    #
    # port configuration
    #
    my %port = %{ $data{ports}{ $index } };
    my $db_name      = "";
    my $port_id      = "";
    my $port_alias   = "";
    my $port_type    = "";
    my $port_subtype = "";

    if ( defined $port{ifName} ) {
      $port_id = name_format( $port{ifName} );
      $db_name = $port_id . ".rrd";
    }
    else { next; }

    if ( defined $port{ifType} ) {
      $port_type = port_type( $port{ifName}, $port{ifType} );

      unless ( $port_type ) { next; }
    }
    else { next; }

    if ( $lan_type eq 'CISCO' && $data{fex_ports}{ $index } ) {
      $port_subtype = 'fexport';
      $conf{ports}{ $index }{port_subtype} = $port_subtype;
    }

    $conf{ports}{ $index }{port_id}   = $port_id;
    $conf{ports}{ $index }{port_type} = $port_type;

    if ( $port_type eq 'stackport' ) {
      next;
    }

    if ( defined $port{ifAlias} && $port{ifAlias} ne '' ) {
      $port_alias = $port{ifAlias};
    }

    #
    # port performance stats
    #
    my $data_out            = "";
    my $data_in             = "";
    my $unicast_pkts_out    = "";
    my $unicast_pkts_in     = "";
    my $multicast_pkts_out  = "";
    my $multicast_pkts_in   = "";
    my $broadcast_pkts_out  = "";
    my $broadcast_pkts_in   = "";
    my $errors_out          = "";
    my $errors_in           = "";
    my $reserve1            = "";
    my $reserve2            = "";

    if ( isdigit( $port{ifHCOutOctets} ) )      { $data_out           = $port{ifHCOutOctets}; }
    if ( isdigit( $port{ifHCInOctets} ) )       { $data_in            = $port{ifHCInOctets}; }
    if ( isdigit( $port{ifOutUcastPkts} ) )     { $unicast_pkts_out   = $port{ifOutUcastPkts}; }
    if ( isdigit( $port{ifInUcastPkts} ) )      { $unicast_pkts_in    = $port{ifInUcastPkts}; }
  #  if ( isdigit( $port{ifOutMulticastPkts} ) ) { $multicast_pkts_out = $port{ifOutMulticastPkts}; }
  #  if ( isdigit( $port{ifInMulticastPkts} ) )  { $multicast_pkts_in  = $port{ifInMulticastPkts}; }
  #  if ( isdigit( $port{ifOutBroadcastPkts} ) ) { $broadcast_pkts_out = $port{ifOutBroadcastPkts}; }
  #  if ( isdigit( $port{ifInBroadcastPkts} ) )  { $broadcast_pkts_in  = $port{ifInBroadcastPkts}; }
    if ( isdigit( $port{ifOutErrors} ) )        { $errors_out         = $port{ifOutErrors}; }
    if ( isdigit( $port{ifInErrors} ) )         { $errors_in          = $port{ifInErrors}; }

    ## filter VLAN interface with no traffic
    if ( $port_type eq 'vlan' && ( $data_out eq '0' || $data_out eq '' ) && ( $data_in eq '0' || $data_in eq '' )
         && ( $unicast_pkts_out eq '0' || $unicast_pkts_out eq '' ) && ($unicast_pkts_in eq '0' || $unicast_pkts_in eq '') ) {
      next;
    }

    # filter non-used ports
    if ( $data_out ne '0' && $data_out ne '' && $data_in ne '0' && $data_in ne '' ) {
      print $pcfg "$port_id : $db_name : $port_alias : $port_type : $port_subtype\n";
    }

    print OUT "PORT,$act_time,$port_id,$data_out,$data_in,$unicast_pkts_out,$unicast_pkts_in,$multicast_pkts_out,$multicast_pkts_in,$broadcast_pkts_out,$broadcast_pkts_in,$errors_out,$errors_in,$reserve1,$reserve2,$switch_name,$db_name\n";
  } ## end foreach my $name ( sort ... ports )

  #
  # CPU
  #
  print OUT "\n#CPU,act_time,cpu_name,utilization,reserve1,reserve2,switch_name,db_name\n";
  foreach my $index ( sort { $a <=> $b } keys( %{ $data{cpu} } ) ) {
    my %cpu        = %{ $data{cpu}{ $index } };
    my $db_name    = "";
    my $cpu_name   = "";
    my $phys_index = $cpu{cpmCPUTotalPhysicalIndex};

    if ( defined $phys_index && $phys_index eq '0' ) {
      $cpu_name = 'CPU';
    }
    elsif ( defined $phys_index && defined $data{physical_entries}{ $phys_index }{entPhysicalName} ) {
      $cpu_name = $data{physical_entries}{ $phys_index }{entPhysicalName};
      $cpu_name =~ s/\s+/_/g;
      $cpu_name =~ s/\W//g;
      if ( $cpu_name eq '' ) { next; }
    }
    else { next; }

    $db_name = $cpu_name . ".lcd";

    my $utilization = '';

    if ( defined $cpu{cpmCPUTotal5minRev} && $cpu{cpmCPUTotal5minRev} =~ /\d{1,3} percent/ ) {
      $utilization = $cpu{cpmCPUTotal5minRev};
      $utilization =~ s/ percent//;
    }

    print OUT "CPU,$act_time,$cpu_name,$utilization,$switch_name,$db_name\n";
  } ## end foreach my $index ( sort ... cpu )

  #
  # MEMORY
  #
  print OUT "\n#MEMORY,act_time,pool_name,utilization,reserve1,reserve2,switch_name,db_name\n";
  foreach my $index ( sort { $a <=> $b } keys( %{ $data{memory} } ) ) {
    my %mem       = %{ $data{memory}{ $index } };
    my $db_name   = "";
    my $pool_name = "";

    if ( defined $mem{ciscoMemoryPoolName} ) {
      $pool_name = $mem{ciscoMemoryPoolName};
      $pool_name =~ s/\s+/_/g;
      $pool_name =~ s/\W//g;
    }
    else { next; }

    $db_name = $pool_name . ".lmd";

    my $utilization = '';

    if ( defined $mem{ciscoMemoryPoolUsed} && isdigit( $mem{ciscoMemoryPoolUsed} )
         && defined $mem{ciscoMemoryPoolFree} && isdigit( $mem{ciscoMemoryPoolFree} ) ) {
      $utilization = 100 * ( $mem{ciscoMemoryPoolUsed} / ( $mem{ciscoMemoryPoolUsed} + $mem{ciscoMemoryPoolFree} ) );
    }

    print OUT "MEMORY,$act_time,$pool_name,$utilization,$switch_name,$db_name\n";
 } ## end foreach my $index ( sort ... cpu )

  close(OUT);
  close($pcfg);
  rename "$ports_cfg_suffix", "$ports_cfg";

  return 1;
} ## sub save_performance_data

sub collect_configuration {
  my $sess            = shift;
  my $conf_touch_file = "$tmpdir/$lan_ip-snmp-conf.touch";

  ## collect configuration only once per hour
  if ( !-f $conf_touch_file || file_time_diff($conf_touch_file) > 3600 ) {
    my $ret = `touch $conf_touch_file`;

    if ( $ret ) {
      error( "$ret " . __FILE__ . ":" . __LINE__ );
      return 0;
    }
    $update_conf = 1;
  }
  else {
    return 1;
  }

  ## switch configurations
  snmp_collect($sess, 'configuration', 'sysLocation');
  snmp_collect($sess, 'configuration', 'sysUpTime');
  snmp_collect($sess, 'physical_entries', 'entPhysicalDescr');
  snmp_collect($sess, 'physical_entries', 'entPhysicalClass');
  snmp_collect($sess, 'physical_entries', 'entPhysicalName');
  snmp_collect($sess, 'physical_entries', 'entPhysicalSerialNum');
  snmp_collect($sess, 'physical_entries', 'entPhysicalModelName');
  snmp_collect($sess, 'physical_entries', 'entPhysicalSoftwareRev');

  ## port configurations
  snmp_collect($sess, 'ports', 'ifLastChange');
  snmp_collect($sess, 'port_channels', 'dot3adAggPortSelectedAggID');

  ## vlans
  if ( $mib_compatibility{"Q-BRIDGE-MIB"} ) {
    snmp_collect($sess, 'vlans', 'dot1qVlanFdbId', undef, 1);
    snmp_collect($sess, 'vlans', 'dot1qVlanCurrentEgressPorts', undef, 1);
    snmp_collect($sess, 'vlans', 'dot1qVlanCurrentUntaggedPorts', undef, 1);
    snmp_collect($sess, 'vlans', 'dot1qVlanStaticName');
    snmp_collect($sess, 'ports', 'dot1qPvid');
  }

  ## lldp
  if ( $mib_compatibility{"LLDP-MIB"} &&  !$mib_compatibility{"CISCO-CDP-MIB"}) {
    snmp_collect($sess, 'lldp', 'lldpRemChassisIdSubtype');
    snmp_collect($sess, 'lldp', 'lldpRemChassisId');
    snmp_collect($sess, 'lldp', 'lldpRemPortId');
    snmp_collect($sess, 'lldp', 'lldpRemSysName');
    snmp_collect($sess, 'lldp', 'lldpRemSysCapEnabled', \&parse_capabilities_lldp);
  }

  ## macs
  snmp_collect($sess, 'mac_entries', 'dot1dTpFdbAddress', \&mac_format);
  snmp_collect($sess, 'mac_entries', 'dot1dTpFdbPort');
  snmp_collect($sess, 'mac_index', 'dot1dBasePortIfIndex');

  if ( $lan_type eq "CISCO" ) {
    collect_configuration_cisco( $sess );
  }

  return 1;
}

sub save_configuration {
  if ( $update_conf ) {
    if ( keys %{ $data{ports} } ) { ports_conf(); }
    if ( keys %{ $data{vlans} } || keys %{ $data{vlan_member_ports} } ) { vlans(); }
    if ( keys %{ $data{port_channels} } ) { port_channels(); }
    if ( keys %{ $data{lldp} } ) { topology(); }
    if ( $lan_type eq 'CISCO' ) {
      if ( keys %{ $data{fex_config} } ) { fex_cisco(); }
      if ( keys %{ $data{vss_chassis} } ) { vss_cisco(); }
      if ( keys %{ $data{cdp} } ) { topology(); }
      if ( keys %{ $data{stack} } ) { stack_cisco(); }
    }
    # config.cfg
    if ( scalar keys %{ $data{configuration} } ) { switch_conf(); }

    save_as_json( "$wrkdir/$switch_name/config.cfg", \%conf);
  }

  return 1;
} ## end sub save_configurations

sub ports_conf {
  my $up_time = $data{configuration}{sysUpTime};

  read_alias_cfg();

  ## ports conf
  foreach my $index ( keys %{ $data{ports} } ) {
    my %port           = %{ $data{ports}{ $index } };
    my $port_name      = $port{ifName};
    my $speed          = $port{ifHighSpeed};
    my $alias_sw       = $port{ifAlias};
    my $alias_cfg      = $port{cfgAlias};
    my $operStatus     = $port{ifOperStatus};
    my $last_change    = $port{ifLastChange};
    my $admin_status   = $port{ifAdminStatus};

    if ( defined $port_name && defined $conf{ports}{ $index }{port_type} ) {
      $conf{ports}{ $index }{port_name} = $port_name;
    }
    else {
      next;
    }

    if ( defined $speed )        { $conf{ports}{ $index }{speed} = $speed;}
    if ( defined $alias_sw )     { $conf{ports}{ $index }{alias_from_sw} = $alias_sw; }
    if ( defined $alias_cfg )    { $conf{ports}{ $index }{alias_from_cfg} = $alias_cfg; }
    if ( defined $alias_sw )     { $conf{ports}{ $index }{alias_from_sw} = $alias_sw; }
    if ( defined $admin_status ) { $conf{ports}{ $index }{admin_status} = $admin_status; }

    if ( defined $operStatus ) {
      $conf{ports}{ $index }{state} = $operStatus eq 'up' ? "used" : "unused"; }
    elsif ( defined $last_change && defined $up_time ) {
      $conf{ports}{ $index }{state} = happened_before( $up_time, $last_change ) ? "used" : "unused";
    }

    if ( $lan_type eq 'CISCO' ) {
      ports_conf_cisco( $index, \%port );
    }
  }

  ## mac addresses
  foreach my $key ( keys %{ $data{mac_entries} } ) {
    my $mac_address = $data{mac_entries}{ $key }{dot1dTpFdbAddress};
    my $bridge_port = $data{mac_entries}{ $key }{dot1dTpFdbPort};
    unless ( defined $mac_address && defined $bridge_port ) { next; };

    my $port_index  = $data{mac_index}{ $bridge_port }{dot1dBasePortIfIndex};
    unless ( defined $port_index ) { next; };

    my $port_name   = $data{ports}{$port_index}{ifName};
    unless ( defined $port_name ) { next; };

    push @{ $conf{ports}{ $port_index }{macs} }, $mac_address;
  }

  return 1;
} ## end sub ports_conf

sub topology {
  foreach my $index ( keys %{ $data{lldp} } ) {
    my %lldp_entry   = %{ $data{lldp}{$index} };
    my @capabilities = defined $lldp_entry{lldpRemSysCapEnabled} ? @{ $lldp_entry{lldpRemSysCapEnabled} } : ();
    my $if_index     = ( split /\./, $index )[1];

    unless ( defined $lldp_entry{lldpRemSysCapEnabled} && scalar grep(/^bridge$/, @{ $lldp_entry{lldpRemSysCapEnabled} } ) ) {
      next;
    }

    if ( defined $lldp_entry{lldpRemChassisIdSubtype} ) {
      $conf{topology}{ $if_index }{address_type} = $lldp_entry{lldpRemChassisIdSubtype};
    }
    if ( defined $lldp_entry{lldpRemChassisId} ) {
      $conf{topology}{ $if_index }{address} = $lldp_entry{lldpRemChassisId};
    }
    if ( defined $lldp_entry{lldpRemPortId} ) {
      $conf{topology}{ $if_index }{remote_port} = $lldp_entry{lldpRemPortId};
    }
    if ( defined $lldp_entry{lldpRemSysName} ) {
      $conf{topology}{ $if_index }{remote_system} = $lldp_entry{lldpRemSysName};
    }
  }

  foreach my $index ( keys %{ $data{cdp} } ) {
    my %cdp_entry   = %{ $data{cdp}{$index} };

    if ( $cdp_entry{cdpCacheAddressType} ) {
      $conf{topology}{ $index }{address_type} = $cdp_entry{cdpCacheAddressType};
    }
    if ( $cdp_entry{cdpCacheAddress} ) {
      $conf{topology}{ $index }{address}      = $cdp_entry{cdpCacheAddress};
    }
    if ( $cdp_entry{cdpCacheDevicePort} ) {
      $conf{topology}{ $index }{remote_port}  = $cdp_entry{cdpCacheDevicePort};
    }

    if ( defined $cdp_entry{cdpCacheSysName} ) {
      $conf{topology}{ $index }{remote_system} = $cdp_entry{cdpCacheSysName};
    }
    elsif ( defined $cdp_entry{cdpCacheDeviceId} ) {
      $conf{topology}{ $index }{remote_system} = $cdp_entry{cdpCacheDeviceId};
    }
  }

  return 1;
} ## end sub topology

sub vlans {
  ## get vlans from indexed BRIDGE-MIB
  if ( $lan_type eq 'CISCO' && keys %{ $data{vlan_member_ports} } ) {
    vlans_cisco();
    return 1;
  }

  ## non-indexed vlans
  ## pvid
  foreach my $index (keys %{ $conf{ports} }) {
    my $pvid = $data{ports}{ $index }{dot1qPvid};

    if ( defined $pvid ) {
      push @{ $conf{ports}{ $index }{vlans}{ $pvid } }, "P";
    }
  }

  ## tagged vs untagged vlans
  while ( my ($key, $vlan) = each %{ $data{vlans} } ) {
    my %vlan              = %{ $data{vlans}{ $key } };
    my $vlan_name         = $vlan{dot1qVlanStaticName};
    my $portlist_egress   = $vlan{dot1qVlanCurrentEgressPorts};
    my $portlist_untagged = $vlan{dot1qVlanCurrentUntaggedPorts};

    unless ( defined $portlist_egress && defined $portlist_untagged ) {
      next;
    }

    if ( defined $vlan_name ) {
      $conf{vlans}{ $key }{vlan_name} = $vlan_name;
    }

    my @list_egress   = @{ parse_portlist( $portlist_egress ) };
    my @list_untagged = @{ parse_portlist( $portlist_untagged ) };

    for my $index (@list_egress) { #either tagged or untagged
      my $port_name = $conf{ports}{ $index }{port_name};
      my $vlan_id   = $vlan{dot1qVlanFdbId};

      unless ( defined $port_name && defined $vlan_id ) { next; }
      push @{ $conf{ports}{ $index }{vlans}{ $vlan_id } }, "T";
    }

    for my $index (@list_untagged) {
      my $port_name = $conf{ports}{ $index }{port_name};
      my $vlan_id   = $vlan{dot1qVlanFdbId};

      unless ( defined $port_name && defined $vlan_id ) { next; }
      push @{ $conf{ports}{ $index }{vlans}{ $vlan_id } }, "E";
    }
  }

  return 1;
} ## end sub vlans

sub port_channels {
  foreach my $index ( keys %{ $data{port_channels} } ) {
    my $port_channel;
    my $agg_index  = $data{port_channels}{ $index }{dot3adAggPortSelectedAggID};
    my $pagp_index = $data{port_channels}{ $index }{pagpGroupIfIndex};

    if ( defined $agg_index && $agg_index ne $index && $agg_index ne '0' && defined $data{ports}{ $index } ) {
      $port_channel = $agg_index;
    }
    elsif ( defined $pagp_index && $pagp_index ne $index && $pagp_index ne '0' && defined $data{ports}{ $index } ) {
      $port_channel = $pagp_index;
    }
    else {
      next;
    }

    my %member_port = %{ $data{ports}{ $index } };
    if ( defined $member_port{ifName} ) {
      push @{ $conf{ports}{ $port_channel }{port_channel_members} }, $member_port{ifName};
      push @{ $conf{ports}{ $port_channel }{port_channel_indexes} }, $index;
    }
  }

  return 1;
} ## end sub port_channels

sub switch_conf {
  my $switch_ip              = $dest;
  my $switch_location        = defined $data{configuration}{sysLocation} ? $data{configuration}{sysLocation} : "";
  my $switch_model           = "";
  my $switch_os_version      = "";
  my $switch_serial_number   = "";
  my $switch_days_up         = defined $data{configuration}{sysUpTime} ? $data{configuration}{sysUpTime} : "";
  my $switch_total_ports     = 0;
  my $switch_unused_ports    = 0;
  my $switch_total_isl_ports = 0;

  ## Days up
  if ( $switch_days_up ) {
    $switch_days_up = (split ':', $switch_days_up)[0];
  }

  ## Port count, unused port count
  while ( my ($index, $port_id) = each %{ $conf{ports} } ) {
    my %port = %{ $conf{ports}{ $index } };
    if ( defined $conf{ports}{ $index }{port_type} && $conf{ports}{ $index }{port_type} eq 'physical' ) {
      $switch_total_ports++;
      $switch_unused_ports++ if ( defined $port{state} && $port{state} eq 'unused' );
    }
  }

  ## Serial number, os version, model
  if ( scalar keys %{ $data{physical_entries} } ) {
    foreach my $phys_entry_index ( sort keys %{ $data{physical_entries} } ) {
      my %phys_entry = %{ $data{physical_entries}{ $phys_entry_index } };

      if ( defined $phys_entry{entPhysicalClass} && $phys_entry{entPhysicalClass} eq 'chassis'
           && defined $phys_entry{entPhysicalSerialNum} && defined $phys_entry{entPhysicalModelName}
           && $switch_model eq '' && $switch_serial_number eq '' ) {
        $switch_serial_number = $phys_entry{entPhysicalSerialNum};
        $switch_model         = $phys_entry{entPhysicalModelName};

        if ( defined $phys_entry{entPhysicalSoftwareRev} ) {
          $switch_os_version = $phys_entry{entPhysicalSoftwareRev};
          last;
        }
      }
      elsif ( $switch_model ne '' && $switch_serial_number ne '' && defined $phys_entry{entPhysicalSoftwareRev}
           && defined $phys_entry{entPhysicalClass} && $phys_entry{entPhysicalClass} eq 'module' ) {
        $switch_os_version = $phys_entry{entPhysicalSoftwareRev};
        last;
      }
    }
  }

  ## ISL count
  $switch_total_isl_ports = scalar keys %{ $conf{topology} };

  $conf{switch}{switch_name}            = $switch_name;
  $conf{switch}{switch_ip}              = $switch_ip;
  $conf{switch}{switch_location}        = $switch_location;
  $conf{switch}{switch_model}           = $switch_model;
  $conf{switch}{switch_os_version}      = $switch_os_version;
  $conf{switch}{switch_serial_number}   = $switch_serial_number;
  $conf{switch}{switch_days_up}         = $switch_days_up;
  $conf{switch}{switch_total_ports}     = $switch_total_ports;
  $conf{switch}{switch_unused_ports}    = $switch_unused_ports;
  $conf{switch}{switch_total_isl_ports} = $switch_total_isl_ports;

  return 1;
} ## end sub switch_conf

sub save_health_status {
  my $health_status_txt           = "$wrkdir/$switch_name/HEALTH_STATUS.txt";
  my $health_status_txt_time_diff = file_time_diff($health_status_txt);
  my @health_status_data;

  # global health status section
  my $main_status    = "OK";
  my $state_suffix  = "ok";
  my $act_timestamp = time();
  my @error_message = ();
  my @states;

  foreach my $index ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
    my $port_id          = $conf{ports}{$index}{port_id};
    my $port_type        = $conf{ports}{$index}{port_type};
    my $admin_status     = "";
    my $oper_status      = "";

    unless ( defined $port_id && defined $port_type && $port_type eq 'physical' ) { next; }

    if ( defined $data{ports}{$index}{ifAdminStatus} ) {
      $admin_status = $data{ports}{$index}{ifAdminStatus};
    }
    if ( defined $data{ports}{$index}{ifOperStatus} ) {
      $oper_status = $data{ports}{$index}{ifOperStatus};
    }

    if ( defined $oper_status && $oper_status eq "lowerLayerDown" ) {
      $main_status   = "NOT_OK";
      $state_suffix = "nok";
      push(@error_message, "The operational status of the port $port_id is $oper_status.\n");
    }

    push( @health_status_data, "$port_id,$admin_status,$oper_status\n" );
  }

  if ( !-f $health_status_txt || $health_status_txt_time_diff > 299 ) {
    open( HS, ">$health_status_txt-tmp" ) || error( "Couldn't open file $health_status_txt-tmp $!" . __FILE__ . ":" . __LINE__ ) && exit;
    print HS "port,Admin status,Operational status\n";
    print HS "-------------------------------------\n";
    print HS @health_status_data;
    close(HS);
    rename( "$health_status_txt-tmp", "$health_status_txt" );
  } ## end if ( !-f $health_status_txt...)

  if ( $lan_type eq 'CISCO' ) {
    if ( modules_status_cisco() eq 'NOT_OK' ) {
      $main_status   = 'NOT_OK';
      $state_suffix  = "nok";
    }
  }

  my $alert_status   = 0;
  my $component_name = $switch_name;
  my $hs_summary_dir = "$tmpdir/health_status_summary";
  $component_name =~ s/\s+//g;

  if ( !-d $hs_summary_dir ) {
    mkdir( $hs_summary_dir, 0755 ) || error( "Cannot mkdir $hs_summary_dir: $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }

  if ( !-f "$hs_summary_dir/$component_name.ok" && !-f "$hs_summary_dir/$component_name.nok" ) {
    $alert_status = 1;
  }
  elsif ( -f "$hs_summary_dir/$component_name.ok" ) {
    unlink("$hs_summary_dir/$component_name.ok");
    $alert_status = 1;
  }
  elsif ( -f "$hs_summary_dir/$component_name.nok" ) {
    unlink("$hs_summary_dir/$component_name.nok");
  }

  if ( $state_suffix eq "nok" && $alert_status == 1 ) {
    ### alert health status
    if ( exists &Xorux_lib::health_status_alerting ) {
      Xorux_lib::health_status_alerting( $component_name, \@error_message);
    }
  } ## end if ( $state_suffix eq ...)

  open( MAST, '>', "$hs_summary_dir/$component_name.$state_suffix" ) || error( "Couldn't open file $hs_summary_dir/$component_name.$state_suffix $!" . __FILE__ . ":" . __LINE__ ) && exit;
  print MAST "LAN-SWITCH : $switch_name : $main_status : $act_timestamp\n";
  close(MAST);

  return 1;
} ## end sub save_health_status


###########
#  CISCO  #
###########

sub collect_configuration_cisco {
  my $sess = shift;

  if ( $mib_compatibility{"CISCO-PAGP-MIB"} ) {
    snmp_collect($sess, 'port_channels', 'pagpGroupIfIndex');
  }

  ## vlan
  if ( $mib_compatibility{"CISCO-VTP-MIB"} ) {
    snmp_collect($sess, 'vlan_info', 'vtpVlanName', undef, 1);
    snmp_collect($sess, 'vlan_info', 'vtpVlanState', undef, 1);
    snmp_collect($sess, 'vlan_info', 'vtpVlanType', undef, 1);
    snmp_collect($sess, 'vlan_info', 'vtpVlanIfIndex', undef, 1);
  }

  ## routed vlan-iftable relationship
  if ( $mib_compatibility{"CISCO-VLAN-IFTABLE-RELATIONSHIP-MIB"} ) {
    snmp_collect($sess, 'vlan_info', 'cviRoutedVlanIfIndex', undef, 0);
  }

  ## stackport
  if ( $mib_compatibility{"CISCO-STACKWISE-MIB"} ) {
    snmp_collect($sess, 'ports', 'cswStackPortNeighbor');
    snmp_collect($sess, 'ports', 'cswStackPortOperStatus');
    snmp_collect($sess, 'stack', 'cswSwitchRole');
    snmp_collect($sess, 'stack', 'cswSwitchMacAddress');
  }

  ## FEX
  if ( $mib_compatibility{"CISCO-ETHERNET-FABRIC-EXTENDER-MIB"} ) {
    snmp_collect($sess, 'fex_ports', 'cefexBindingExtenderIndex');
    snmp_collect($sess, 'fex_config', 'cefexConfigExtenderName');
  }

  ## VSS
  if ( $mib_compatibility{"CISCO-VIRTUAL-SWITCH-MIB"} ) {
    snmp_collect($sess, 'vss_chassis', 'cvsChassisRole');
    snmp_collect($sess, 'vss_chassis', 'cvsChassisUpTime');
    snmp_collect($sess, 'vss_chassis', 'cvsChassisSwitchID');
    snmp_collect($sess, 'vss_ports', 'cvsVSLConnectOperStatus');
    snmp_collect($sess, 'vss_ports', 'cvsVSLCoreSwitchID');
  }

  ## topology, neighbours
  if ( $mib_compatibility{"CISCO-CDP-MIB"} ) {
    snmp_collect($sess, 'cdp', 'cdpCacheAddressType', undef, 0);
    snmp_collect($sess, 'cdp', 'cdpCacheAddress', \&ip_decimal_format, 0);
    snmp_collect($sess, 'cdp', 'cdpCacheDeviceId', undef, 0);
    snmp_collect($sess, 'cdp', 'cdpCacheDevicePort', undef, 0);
    snmp_collect($sess, 'cdp', 'cdpCacheCapabilities', \&parse_capabilities_cdp, 0);
    snmp_collect($sess, 'cdp', 'cdpCacheSysName', undef, 0);
  }

  ## macs and vlans
  if ( $mib_compatibility{"COMMUNITY-STRING-INDEX"} && keys %{ $data{vlan_info} } ) {
    foreach my $vlan ( keys %{ $data{vlan_info} } ) {
      my %vlan_entry = %{ $data{vlan_info}{ $vlan } };
      unless ( defined $vlan_entry{vtpVlanType} && $vlan_entry{vtpVlanType} eq 'ethernet'
               && defined $vlan_entry{vtpVlanState} && $vlan_entry{vtpVlanState} eq 'operational' ) {
        next;
      }

      my $indexed_sess = snmp_session( $vlan );
      snmp_collect($indexed_sess, 'mac_entries', 'dot1dTpFdbAddress', \&mac_format);
      snmp_collect($indexed_sess, 'mac_entries', 'dot1dTpFdbPort');
      snmp_collect($indexed_sess, 'mac_index', 'dot1dBasePortIfIndex');

      my %if_indexes = %{ snmp_collect($indexed_sess, undef, 'dot1dBasePortIfIndex') };
      push @{ $data{vlan_member_ports}{ $vlan } }, values %if_indexes;
    }
  }

  return 1;
} ## end sub collect_configuration_cisco

sub ports_conf_cisco {
  my $index              = shift;
  my %port               = %{ shift @_ };
  my $stackport_status   = $port{cswStackPortOperStatus};
  my $stackport_neighbor = $port{cswStackPortNeighbor};

  if ( defined $stackport_status ) {
    $conf{ports}{ $index }{stackport_status} = $stackport_status;
  }

  if ( defined $stackport_neighbor && $stackport_neighbor ne '0' ) {
    if ( defined $data{physical_entries}{ $stackport_neighbor }{entPhysicalSerialNum} ) {
      $conf{ports}{ $index }{stack_neighbor_serial} = $data{physical_entries}{ $stackport_neighbor }{entPhysicalSerialNum};
    }

    if ( defined $data{stack}{ $stackport_neighbor }{cswSwitchMacAddress} ) {
      $conf{ports}{ $index }{stack_neighbor_mac} = $data{stack}{ $stackport_neighbor }{cswSwitchMacAddress};
    }

    if ( defined $data{stack}{ $stackport_neighbor }{cswSwitchRole} ) {
      $conf{ports}{ $index }{stack_neighbor_role} = $data{stack}{ $stackport_neighbor }{cswSwitchRole};
    }
  }

  return 1;
} ## end sub ports_conf_cisco

sub fex_cisco {
  foreach my $index ( keys %{ $data{fex_ports} } ) {
    my $fex_binding = $data{fex_ports}{ $index }{cefexBindingExtenderIndex};

    if ( defined $fex_binding && defined $data{ports}{ $index }{ifName} ) {
      push @{ $conf{fex}{ $fex_binding }{fex_connection} }, $data{ports}{ $index }{ifName};
    }
  }

  foreach my $index ( keys %{ $data{fex_config} } ) {
    my $fex_name = $data{fex_config}{ $index }{cefexConfigExtenderName};

    if ( defined $fex_name ) {
      $conf{fex}{ $index }{fex_name} = $fex_name;
    }

    foreach my $phys_index ( keys %{ $data{physical_entries} } ) {
      my %phys_entry = %{ $data{physical_entries}{ $phys_index } };

      if ( defined $phys_entry{entPhysicalClass} && $phys_entry{entPhysicalClass} eq 'chassis'
           && defined $phys_entry{entPhysicalDescr} && $phys_entry{entPhysicalDescr} =~ /^Fex-$index/
           && $phys_index =~ /$index$/ ) {

        if ( defined $phys_entry{entPhysicalSerialNum} ) {
          $conf{fex}{ $index }{fex_serial} = $phys_entry{entPhysicalSerialNum};
        }

        if ( defined $phys_entry{entPhysicalModelName} ) {
          $conf{fex}{ $index }{fex_model} = $phys_entry{entPhysicalModelName};
        }
      }
    }
  }

  return 1;
} ## end sub fex_cisco

sub vlans_cisco {
  while ( my ($vlan_id, $ports) = each %{ $data{vlan_member_ports} } ) {
    for my $index ( @{ $ports } ) {
      push @{ $conf{ports}{ $index }{vlans}{ $vlan_id } }, ""; # no inforamtion about tagged/untagged
    }
  }

  foreach my $index ( keys %{ $data{vlan_info} } ) {
    my %vlan_entry       = %{ $data{vlan_info}{ $index } };
    my $vlan_name        = $vlan_entry{vtpVlanName};
    my $unrouted_ifindex = $vlan_entry{vtpVlanIfIndex};
    my $routed_ifindex   = $vlan_entry{cviRoutedVlanIfIndex};

    if ( defined $vlan_name ) {
      $conf{vlans}{ $index }{vlan_name} = $vlan_name;
    }
    else {
      next;
    }

    if ( defined $unrouted_ifindex && $unrouted_ifindex ne '0' ) {
      my %interface = %{ $data{ports}{ $unrouted_ifindex } };

      if ( defined $interface{ifName} )    { $conf{vlans}{ $index }{unrouted_interface} = $interface{ifName}; }
      if ( defined $interface{ifAlias} )   { $conf{vlans}{ $index }{unrouted_ifalias}   = $interface{ifAlias}; }
      if ( defined $interface{cfgAlias} )  { $conf{vlans}{ $index }{unrouted_cfgalias}  = $interface{cfgAlias}; }
    }

    if ( defined $routed_ifindex ) {
      my %interface = %{ $data{ports}{ $routed_ifindex } };

      if ( defined $interface{ifName} )    { $conf{vlans}{ $index }{routed_interface} = $interface{ifName}; }
      if ( defined $interface{ifAlias} )   { $conf{vlans}{ $index }{routed_ifalias}   = $interface{ifAlias}; }
      if ( defined $interface{cfgAlias} )  { $conf{vlans}{ $index }{routed_cfgalias}  = $interface{cfgAlias}; }
    }
  }

  return 1;
} ## end sub vlans_cisco

sub stack_cisco {
  foreach my $index ( keys %{ $data{stack} } ) {
    if ( defined $data{stack}{ $index }{cswSwitchRole} ) {
      $conf{stack}{ $index }{role} = $data{stack}{ $index }{cswSwitchRole};
    }

    if ( defined $data{physical_entries}{ $index } ) {
      my %phys_entry = %{ $data{physical_entries}{ $index } };

      if ( defined $phys_entry{entPhysicalSerialNum} ) {
        $conf{stack}{ $index }{serial_number} = $phys_entry{entPhysicalSerialNum};
      }

      if ( defined $phys_entry{entPhysicalModelName} ) {
        $conf{stack}{ $index }{model_number} = $phys_entry{entPhysicalModelName};
      }
    }
  }

  return 1;
} ## end sub stack_cisco

sub vss_cisco {
  foreach my $index ( keys %{ $data{vss_ports} } ) {
    my $switch_id  = $data{vss_ports}{ $index }{cvsVSLCoreSwitchID};
    my $vsl_status = $data{vss_ports}{ $index }{cvsVSLConnectOperStatus};

    unless ( defined $switch_id ) {
      next;
    }

    if ( defined $data{ports}{ $index }{ifName} ) {
      $conf{vss}{ $switch_id }{vss_vsl} = $data{ports}{ $index }{ifName};
    }

    if ( defined $vsl_status ) {
      $conf{vss}{ $switch_id }{vss_vsl_status} = $vsl_status;
    }
  }

  foreach my $index ( keys %{ $data{vss_chassis} } ) {
    my $switch_id      = $data{vss_chassis}{ $index }{cvsChassisSwitchID};
    my $switch_role    = $data{vss_chassis}{ $index }{cvsChassisRole};
    my $switch_uptime  = $data{vss_chassis}{ $index }{cvsChassisUpTime};

    unless ( defined $switch_id ) {
      next;
    }

    if ( defined $switch_uptime ) {
      $conf{vss}{ $switch_id }{vss_uptime} = (split ':', $switch_uptime)[0];
    }

    if ( defined $switch_role ) {
      $conf{vss}{ $switch_id }{vss_role} = $switch_role;
    }

    if ( $data{physical_entries}{ $index } ) {
      my %physical_entry = %{ $data{physical_entries}{ $index } };

      if ( $physical_entry{entPhysicalSerialNum} ) {
        $conf{vss}{ $switch_id }{vss_serial_number} = $physical_entry{entPhysicalSerialNum};
      }

      if ( $physical_entry{entPhysicalDescr} ) {
        $conf{vss}{ $switch_id }{vss_description} = $physical_entry{entPhysicalDescr};
      }

      if ( $physical_entry{entPhysicalName} ) {
        $conf{vss}{ $switch_id }{vss_name} = $physical_entry{entPhysicalName};
      }
    }
  }

  return 1;
} ## end sub vss_cisco

sub modules_status_cisco {
  my $switch_hs_file  = "$wrkdir/$switch_name/HEALTH_STATUS_SWITCH.txt";
  my @failed_statuses = ('failed', 'missing', 'mismatchWithParent', 'mismatchConfig','diagFailed', 'dormant',
                        'outOfServiceAdmin','outOfServiceEnvTemp','poweredDown', 'poweredUp', 'powerDenied',
                        'powerCycled', 'okButPowerOverWarning', 'okButPowerOverCritical', 'syncInProgress',
                        'upgrading', 'okButAuthFailed', 'mdr');
  my $main_status     = 'OK';
  my %switch_status;

  unless ( scalar keys %{ $data{switch_status} } ) {
    if ( -f $switch_hs_file ) {
      unlink($switch_hs_file);
    }

    return 1;
  }

  foreach my $phys_index ( keys %{ $data{switch_status} } ) {
    my $modul_name   = $data{physical_entries}{ $phys_index }{entPhysicalDescr};
    my $modul_serial = $data{physical_entries}{ $phys_index }{entPhysicalSerialNum};
    my $admin_status = $data{switch_status}{ $phys_index }{cefcModuleAdminStatus};
    my $oper_status  = $data{switch_status}{ $phys_index }{cefcModuleOperStatus};

    if ( defined $modul_name && defined $modul_serial && defined $admin_status && defined $oper_status ) {
      $switch_status{modules}{ $phys_index }{modul_name}   = $modul_name;
      $switch_status{modules}{ $phys_index }{modul_serial} = $modul_serial;
      $switch_status{modules}{ $phys_index }{admin_status} = $admin_status;
      $switch_status{modules}{ $phys_index }{oper_status}  = $oper_status;

      foreach my $status ( @failed_statuses ) {
        if ( $oper_status eq $status ) {
          $main_status = 'NOT_OK';
        }
      }
    }
  }

  if ( scalar keys %switch_status ) {
    $switch_status{overall_status} = $main_status;
    save_as_json($switch_hs_file, \%switch_status);
  }

  return $main_status;
} ## end sub modules_status_cisco

sub save_as_json {
  my $path        = shift;
  my $hash_ref    = shift;
  my $path_suffix = "$path-tmp";
  my $json        = JSON->new;

  $json->pretty([1]);
  $json->canonical([1]);

  open( FILE, ">", "$path_suffix" ) || error( "Couldn't open file $path_suffix $!" . __FILE__ . ":" . __LINE__ ) && exit;
  print FILE $json->encode( $hash_ref );
  close(FILE);
  rename($path_suffix, $path);

  return 1;
} ## end sub save_as_json

sub read_json {
  my $file = shift;
  my $json = JSON->new;
  my $string;
  my $output;

  {
    local $/=undef;
    open ( FILE, '<', $file ) or return 0;
    $string = <FILE>;
    close FILE;
  }

  if ($string) {
    eval {
      $output = $json->decode($string);
      1;
    } or do {
      error("Couldn't parse json file $file $@ " . __FILE__ . ":" . __LINE__) && return 0;
    };
  }

  return $output;
} ## end sub read_json

sub read_alias_cfg {
  my $alias_file  = "$basedir/etc/alias.cfg";
  my @alias_lines;

  #read alias.cfg
  if ( -f $alias_file ) {
    open( FHA, "< $alias_file" ) || error( "Couldn't open file $alias_file $!" . __FILE__ . ":" . __LINE__ ) && return 0;
    @alias_lines = <FHA>;
    close(FHA);
  }

  foreach my $line ( @alias_lines ) {
    # alias.cfg entries format: LANPORT:cisco-sg200-50:gi1:test
    if ( $line =~ /^LANPORT:([^:]*):([^:]*):([^:]*)$/ && $1 eq $switch_name && $2 ne '' && $3 ne '') {
      foreach my $index ( keys %{ $data{ports} } ) {
        if ( $data{ports}{ $index }{ifName} eq $2 ) {
          $data{ports}{ $index }{cfgAlias} = $3;
        }
      }
    }
  }

  return 1;
} ## end sub read_alias_cfg

sub file_time_diff {
  my $file = shift;

  my $act_time  = time();
  my $file_time = $act_time;
  my $time_diff = 0;

  if ( -f $file ) {
    $file_time = ( stat($file) )[9];
    $time_diff = $act_time - $file_time;
  }

  return ($time_diff);
} ## end sub file_time_diff

sub isdigit {
  my $digit = shift;

  if ( !defined($digit) || $digit eq '' ) {
    return 0;
  }

  my $digit_work = $digit;
  $digit_work =~ s/[0-9]//g;
  $digit_work =~ s/\.//;
  $digit_work =~ s/^-//;
  $digit_work =~ s/^\+//;
  $digit_work =~ s/e\+//;
  $digit_work =~ s/e\-//;

  if ( length($digit_work) == 0 ) {

    # is a number
    return 1;
  }

  # NOT a number
  return 0;
} ## end sub isdigit

# error handling
sub error {
  my $text     = shift;
  my $act_time = localtime();
  chomp($text);

  #print "ERROR          : $text : $!\n";
  print STDERR "$act_time: $text : $!\n";

  #print "$act_time: $text : $!\n" if $DEBUG > 2;;

  return 1;
} ## end sub error

sub hex2dec {
  my $num = shift;

  $num =~ s/\s+//g;
  my $hex = "0x$num";

  use Math::BigInt;
  my $dec = Math::BigInt->new($hex);

  return ($dec);
} ## end sub hex2dec

sub bignum {
  my $bnum = shift;

  use Math::BigInt;
  my $num = Math::BigInt->new($bnum);

  return ($num);
} ## end sub bignum

sub parse_portlist {
  my $portlist = shift;
  my @parts = split ' ', $portlist;
  my $i = 1;
  my @list_of_ports;

  for my $part (@parts) {
    my $dec = hex( $part );
    my @indexes = $i .. $i + 8;

    push @list_of_ports, map { $dec & ( 128 >> $_ ) ? $indexes[$_] : () } 0 .. 7;
    $i += 8;
  }
  return \@list_of_ports;
} ## end sub parse_portlist

sub ip_decimal_format {
  my $input = shift;
  return $input =~ /([0-9a-fA-F]{2} ?){4}/
    ? join ".", map { hex $_ } split ' ', $input
    : $input;
} ## end sub ip_decimal_format

sub parse_capabilities_cdp {
  my $input = shift;
  $input =~ s/ //g;
  $input = hex( $input );
  my @capabilities = ('Router', 'TB Bridge','SR Bridge','Switch','Host','IGMP conditional filtering','Repeater','VoIP Phone','Remotely-Managed Device');
  my @output;

  for (my $i = 0; $i < scalar @capabilities ; $i++) {
    if ($input & 1<<$i) {
      push @output, $capabilities[$i];
    }
  }
  return \@output;
} ## end sub parse_capabilities_cdp

sub parse_capabilities_lldp {
  my $input = shift;
  $input =~ s/ //g;
  my $dec = hex( $input );
  my @capabilities = ( 'other', 'repeater', 'bridge', 'wlanAccessPoint', 'router', 'telephone', 'docsisCableDevice', 'stationOnly');
  my @output;

  my $mask = (16 ** ((length $input))) >> 1;
  for (my $i = 0; $i < scalar @capabilities; $i++) {
    if ($dec & $mask) {
      push @output, $capabilities[$i];
    }

    $mask >>= 1;
  }

  return \@output;
} ## end sub parse_capabilities lldp

sub mac_format {
  my $input = shift;
  chomp $input;
  $input =~ s/ /:/;

  if ( $input =~ /^(([0-9a-fA-f]{1,2})(:|$)){6}/ ) {
    return join ":", map { length $_ > 1 ? $_ : "0$_" } split /:/, $input;
  }
  else {
    return $input;
  }
} ## end sub mac_format

sub name_format {
  my $name = shift;
  $name =~ s/\//-/g;
  $name =~ s/\s+/_/g;

  return $name;
}

sub port_type {
  my $port_name = shift;
  my $port_type = shift;

  if ( $port_type eq 'ethernetCsmacd' ) { return 'physical'; }
  elsif ( $port_type eq 'tunnel' ) { return 'tunnel'; }
  elsif ( $port_type eq 'propVirtual' || $port_type eq 'ieee8023adLag' ) {
    if ( $port_name =~ /^(VL|Vl|vl)/ ) { return 'vlan'; }
    elsif ( $port_name =~ /^(PO|Po|port-channel)/ ) { return 'port-channel'; }
    elsif ( $port_name =~ /^(Stack|stack)/ ) { return 'stackport'; }
    else { return ''; }
  }
  else { return ''; }
}

sub snmp_collect {
  my $sess               = shift;
  my $section            = shift;
  my $mib                = shift;
  my $transformation_sub = shift;
  my $iid_index          = shift; ## used for wrongly indexed entries with leading or trailing indexes
  my $vb                 = new SNMP::Varbind( [$mib] );
  my %output;

  if ( defined $section && $section eq 'configuration') {
    my $var = $sess->getnext($vb);

    if ( defined $var ) {
      $var =~ s/"//g;
      unless ( $var eq '' ) {
        $data{configuration}{$mib} = $var;
      }
    }
  }
  else {
    for ( my $var = $sess->getnext($vb); ( $vb->tag eq $mib ) and not( $sess->{ErrorNum} ); $var = $sess->getnext($vb) ) {
      my $index = defined $iid_index ? (split /\./, $vb->iid)[$iid_index] : $vb->iid;
      $var =~ s/"//g;
      $var = defined $transformation_sub ? $transformation_sub->($var) : $var;

      if ( $var eq '' ) { next; }

      # get data into hash and return it as reference
      if ( not defined $section ) {
        $output{ $index } = $var;
      }
      # get indexed data into data hash
      else {
        $data{$section}{ $index }{$mib} = $var;
      }
    } ## end for ( $var = $sess->getnext...)

  }
  if ( $sess->{ErrorNum} ) {
    error( "Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib. $!" . __FILE__ . ":" . __LINE__ ) && exit;
    return 0;
  }

  if ( not defined $section ) {
    return \%output;
  }

  return 1;
} ## end sub snmp_collect

sub snmp_test_mib {
  my $sess   = shift;
  my $mib    = shift;
  my $vb     = new SNMP::Varbind( [$mib] );
  my $result = $sess->getnext($vb);

  if ( $vb->tag !~ /\.$mib\./ || $sess->{ErrorNum} ) {
    return 0;
  }

  return 1;
} ## end sub snmp_test_mib

sub happened_before {
  my @a = split(/:|\./, shift);
  my @b = split(/:|\./, shift);

  for (my $i = 0; $i < scalar @a; $i++) {
    if ($a[$i] < $b[$i]) {
      return 1;
    }
    if ($a[$i] > $b[$i]) {
      return 0;
    }
  }

  return 0;
} ## end sub happened_before

sub debug {
  my $ref = shift;

  unless ( defined $ref ) {
    $ref = \%data;
  }

  open( DEBUG, ">", "$wrkdir/$switch_name/debug" ) || error( "Couldn't open file $!" . __FILE__ . ":" . __LINE__ ) && exit;
  print DEBUG Dumper( $ref );
  close(DEBUG);
} ## end sub debug
